"""

In this module, you will find some functions to represent connectivity graphs
and topographic plots over a 2D head model. Enjoy!

@authors: Víctor Martínez-Cagigal and Diego Marcos-Martínez
"""

# External imports
import warnings
import scipy.interpolate as sp
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from matplotlib import cm, colors
import numpy as np

# Medusa imports
from medusa.meeg import UnlocatedChannel


def plot_connectivity(channel_set, adj_mat, head_radius=0.7266,
                      plot_channels=True, plot_skin_in_color=True,
                      plot_clabels=True,plot_contour_ch=False,
                      chcontour_radius=None,interp_points=500,
                      cmap='seismic', show=True, clim=None):

    """ This function depicts a connectivity map over the
        desired channel locations.

        Parameters
        ----------
        channel_set : eeg_standards.EEGChannelSet
            EEG channel set according of class eeg_standards.EEGChannelSet
        adj_mat: numpy.ndarray
            Numpy array with the connectivity values. It must be of the must
            have the following dimensions [n_channels, n_channels]
        head_radius : float
            Head radius. Default is 0.7266, coinciding with FPz. The nasion and
            inion are located at r=1.0
        plot_channels: bool
            Boolean that controls if the channel points should be plotted (default:
            True)
        plot_skin_in_color: bool
            Boolean that controls if the skin of the head should be coloured (default:
            False)
        plot_clabels: bool (Optional)
            Boolean that controls if the channel labels should be plotted (default:
            False)
        plot_contour_ch: bool (Optional)
            Boolean that controls if a contour around each channel should be plotted
            (default: False)
        chcontour_radius: float or None
            Radius of the channel contour if plot_contour_ch is set True. If None
            value, an automatic value is computed, considering the minimum distance
            between channels (default: None)
        interp_points: int (Optional)
            No. interpolation points. The lower N, the lower resolution and faster
            computation (default: 500)
        cmap : str
            Matplotlib colormap
        show : bool
            Show matplotlib figure
        clim : list or None
            Color bar limits. Index 0 contain the lower limit, whereas index 1 must
            contain the upper limit. if none, min and max values are used

        Returns
        -------
        figure : plt.figure
            Figure with the connectivity plot
        """

    # Check adjacency matrix  dimensions
    if adj_mat.shape[0] != len(channel_set.channels):
        raise Exception('Adjacency matrix must have the shape '
                        '[n_channels, n_channels]')

    # Get connectivity values
    values_indx = np.tril_indices(adj_mat.shape[0],1)
    conn_values = adj_mat[values_indx]

    # Map connectivity values to colors
    if clim is None:
        clim = [conn_values.min(),conn_values.max()]
    norm = colors.Normalize(vmin=clim[0],vmax=clim[1],clip=True)
    mapper = cm.ScalarMappable(norm=norm, cmap=cmap)
    conn_colors = mapper.to_rgba(conn_values)

    # Connectivity line widths
    widths = 3 * (np.ones(len(conn_values)) * np.abs(conn_values) - clim[0])/\
             (clim[1] - clim[0])

    ch_x, ch_y = get_cartesian_coordinates(channel_set)

    fig, axes = plot_head(channel_set=channel_set, head_radius=head_radius,
                          plot_channels=plot_channels,
                          plot_skin_in_color=plot_skin_in_color,
                          plot_clabels=plot_clabels,
                          plot_contour_ch=plot_contour_ch,
                          chcontour_radius=chcontour_radius,
                          interp_points=interp_points,
                          show=False)

    edges = []
    for indx, chx in enumerate(ch_x):
        for second_indx in range(indx + 1, len(ch_x)):
            edges.append(
                [[chx, ch_y[indx]], [ch_x[second_indx], ch_y[second_indx]]])

    edges_collection = LineCollection(edges,colors=conn_colors,linewidths=widths)
    axes.add_collection(edges_collection)

    # Depict color bar
    cbar = plt.colorbar(mapper)

    fig.patch.set_alpha(0.0)
    if show is True:
        plt.show(dpi=400)
    return fig, axes


def plot_topography(channel_set, values, head_radius=0.7266,
                    plot_extra=0.29, k=3, make_contour=True, plot_channels=True,
                    plot_skin_in_color=False, plot_clabels=False,
                    plot_contour_ch=False, chcontour_radius=None,
                    interp_points=500, cmap='YlGnBu_r', show=True, clim=None,
                    axes=None, fig=None, show_colorbar=True, linewidth=4.0,
                    background=False):

    """ This function depicts a topographic map of the scalp
    over the desired channel locations.

    Parameters
    ----------
    channel_set : eeg_standards.EEGChannelSet
        EEG channel set according of class eeg_standards.EEGChannelSet
    values: list or numpy.ndarray
        Numpy array with the channel values. It must be of the same size as
        channels.
    head_radius : float
        Head radius. Default is 0.7266, coinciding with FPz. The nasion and
        inion are located at r=1.0
    plot_extra : float
        Extra radius of the plot surface
    k : int
        Number of nearest neighbors for interpolation
    make_contour: bool
        Boolean that controls if the contour lines should be plotted (default:
        True)
    plot_channels: bool
        Boolean that controls if the channel points should be plotted (default:
        True)
    plot_skin_in_color: bool (Optional)
        Boolean that controls if the skin of the head should be coloured (default:
        False)
    plot_clabels: bool (Optional)
        Boolean that controls if the channel labels should be plotted (default:
        False)
    plot_contour_ch: bool (Optional)
        Boolean that controls if a contour around each channel should be plotted
        (default: False)
    chcontour_radius: float or None
        Radius of the channel contour if plot_contour_ch is set True. If None
        value, an automatic value is computed, considering the minimum distance
        between channels (default: None)
    interp_points: int (Optional)
        No. interpolation points. The lower N, the lower resolution and faster
        computation (default: 500)
    cmap : str
        Matplotlib colormap
    background: bool (Optional)
        Set background
    show : bool
        Show matplotlib figure
    axes : matplotlib.pyplot.axes
        If a matplotlib axes are specified, the plot is displayed inside it.
        Otherwise, the plot will generate a new axes.
    clim : list or None
        Color bar limits. Index 0 contain the lower limit, whereas index 1 must
        contain the upper limit. if none, min and max values are used

    Returns
    -------
    figure : plt.figure
        Figure with the topography plot
    """
    # Check values dimensions
    values = np.array(values)
    if values.size != len(channel_set.channels):
        raise Exception('Parameters ch_list and values must have the same '
                        'size')
    if len(values.shape) == 1:
        # Reshape to the correct dimensions [1 x len(ch_list)]
        values = values.reshape(1, -1)
    elif len(values.shape) == 2:
        # Reshape to the correct dimensions [1 x len(ch_list)]
        values = np.squeeze(values).reshape(1, -1)
    else:
        raise Exception('The dimensions of the parameter are not correct')

    # Plot head
    fig, axes = plot_head(channel_set=channel_set,head_radius=head_radius,
                          plot_channels=plot_channels,
                          plot_skin_in_color=plot_skin_in_color,
                          plot_clabels=plot_clabels,
                          plot_contour_ch=plot_contour_ch,
                          chcontour_radius=chcontour_radius,
                          interp_points=interp_points,
                          show=False, axes=axes, fig=fig, linewidth=linewidth,
                          background=background)

    # Create points out of the head to get a natural interpolation
    r_ext_points = 1.5  # Radius of the virtual electrodes
    no_ve = 16  # No. virtual electrodes
    add_x, add_y = pol2cart(r_ext_points * np.ones((1, no_ve)),
                            np.arange(0, 2 * np.pi, 2 * np.pi / no_ve))
    linear_grid = np.linspace(-r_ext_points, r_ext_points, interp_points)
    interp_x, interp_y = np.meshgrid(linear_grid, linear_grid)

    # Get cartesian coordinates
    ch_x, ch_y = get_cartesian_coordinates(channel_set)

    # Create the mask
    mask_radius = np.max(np.sqrt(ch_x**2+ch_y**2)) + plot_extra
    mask = (np.sqrt(np.power(interp_x, 2) +
                    np.power(interp_y, 2)) < mask_radius)

    # Interpolate the data
    ch_x = ch_x.reshape(ch_x.shape[0], 1)
    ch_y = ch_y.reshape(ch_y.shape[0], 1)
    add_values = compute_nearest_values(np.hstack((add_x.T, add_y.T)),
                                        np.hstack((ch_x, ch_y)), values, k)
    grid_points = np.hstack((np.vstack((ch_x, add_x.T)),
                             np.vstack((ch_y, add_y.T))))
    grid_values = np.vstack((values.T, add_values))
    interp_values = np.vstack((interp_x.ravel(), interp_y.ravel())).T
    interp_z = sp.griddata(grid_points, grid_values, interp_values, 'cubic')

    # Mask the data
    interp_z = np.reshape(interp_z, (interp_points, interp_points))
    interp_z[~mask] = float('nan')

    # Plotting the final interpolation
    p_interp = axes.pcolor(interp_x, interp_y, interp_z, cmap=cmap)
    if clim is not None:
        p_interp.set_clim(clim[0], clim[1])
    if show_colorbar:
        cbar = plt.colorbar(p_interp)

    # Plotting the contour
    if make_contour:
        axes.contour(interp_x, interp_y, interp_z, alpha=1, colors='0.2',
                     linewidths=linewidth/4.2)

    if show is True:
        plt.show(dpi=400)
    return fig, axes, p_interp


def plot_head(channel_set, head_radius=0.7266, plot_channels=True,
              plot_skin_in_color=False, plot_clabels=False,
              plot_contour_ch=False, chcontour_radius=None,
              interp_points=500, show=True, axes=None, fig=None,
              linewidth=4.0, background=False):
    """This function depicts a two-dimensional head diagram.

    Parameters
    ----------
    channel_set : eeg_standards.EEGChannelSet
        EEG channel set according of class eeg_standards.EEGChannelSet
    head_radius : float
        Head radius. Default is 0.7266, coinciding with FPz. The nasion and
        inion are located at r=1.0
    plot_channels: bool
        Boolean that controls if the channel points should be plotted (default:
        True)
    plot_skin_in_color: bool (Optional)
        Boolean that controls if the skin of the head should be coloured (default:
        False)
    plot_clabels: bool (Optional)
        Boolean that controls if the channel labels should be plotted (default:
        False)
    plot_contour_ch: bool (Optional)
        Boolean that controls if a contour around each channel should be plotted
        (default: False)
    chcontour_radius: float or None
        Radius of the channel contour if plot_contour_ch is set True. If None
        value, an automatic value is computed, considering the minimum distance
        between channels (default: None)
    interp_points: int (Optional)
        No. interpolation points. The lower N, the lower resolution and faster
        computation (default: 500)
    background: bool (Optional)
        Set background
    show : bool
        Show matplotlib figure
    axes : matplotlib.pyplot.axes
        If a matplotlib axes are specified, the plot is displayed inside it.
        Otherwise, the plot will generate a new figure.
    Returns
    -------
    figure : plt.figure
        Figure with the head plot
    """
    # Check channels errors
    if channel_set.dim != '2D':
        raise ValueError('The channel set must have 2 dimensions')

    # Initialize figure and axis
    if fig is None:
        fig = plt.figure()
    if axes is None:
        axes = fig.add_subplot(111)

    # Compute the cartesian coordinates of each channel
    ch_x, ch_y = get_cartesian_coordinates(channel_set)

    # Plotting the nose
    head_rho = head_radius
    nt = 0.15  # Half-nose width (in percentage of pi/2)
    nr = 0.22  # Nose length (in radius units)
    nose_rho = [head_rho, head_rho + head_rho * nr, head_rho]
    nose_theta = [(np.pi / 2) + (nt * np.pi / 2), np.pi / 2,
                  (np.pi / 2) - (nt * np.pi / 2)]
    nose_x = nose_rho * np.cos(nose_theta)
    nose_y = nose_rho * np.sin(nose_theta)
    axes.plot(nose_x, nose_y, 'k', linewidth=linewidth)
    if plot_skin_in_color:
        axes.fill(nose_x, nose_y, 'k',
                  facecolor='#E8BEAC', edgecolor='k', linewidth=linewidth)

    # Plotting the ears as ellipses
    ellipse_a = 0.08  # Horizontal eccentricity
    ellipse_b = 0.16  # Vertical eccentricity
    ear_angle = 0.9 * np.pi / 8  # Mask angle
    offset = 0.058 * head_radius  # Ear offset
    ear_theta_right = np.linspace(-np.pi / 2 - ear_angle,
                                  np.pi / 2 + ear_angle, interp_points)
    ear_theta_left = np.linspace(np.pi / 2 - ear_angle,
                                 3 * np.pi / 2 + ear_angle, interp_points)
    ear_x_right = ear_rho(ear_theta_right, ellipse_a, ellipse_b) * \
                  np.cos(ear_theta_right)
    ear_y_right = ear_rho(ear_theta_right, ellipse_a, ellipse_b) * \
                  np.sin(ear_theta_right)
    ear_x_left = ear_rho(ear_theta_left, ellipse_a, ellipse_b) * \
                 np.cos(ear_theta_left)
    ear_y_left = ear_rho(ear_theta_left, ellipse_a, ellipse_b) * \
                 np.sin(ear_theta_left)
    axes.plot(ear_x_right + head_rho + offset, ear_y_right, 'k', linewidth=linewidth)
    axes.plot(ear_x_left - head_rho - offset, ear_y_left, 'k', linewidth=linewidth)

    # Plotting the head limits as a circle
    head_theta = np.linspace(0, 2 * np.pi, interp_points)
    head_x = head_rho * np.cos(head_theta)
    head_y = head_rho * np.sin(head_theta)
    axes.plot(head_x, head_y, 'k', linewidth=linewidth)
    if plot_skin_in_color:
        axes.fill(head_x, head_y, facecolor='#E8BEAC',
                  edgecolor='k', linewidth=4)

    if plot_skin_in_color:
        axes.fill(ear_x_right + head_rho + offset, ear_y_right,
                  facecolor='#E8BEAC', edgecolor='k', linewidth=linewidth)
        axes.fill(ear_x_left - head_rho - offset, ear_y_left,
                  facecolor='#E8BEAC', edgecolor='k', linewidth=linewidth)

    # Compute optimal minimum distance between channels
    if chcontour_radius is None:
        dist_matrix = channel_set.compute_dist_matrix()
        dist_matrix.sort()
        min_dist = dist_matrix[:, 1].min()

        #  Adjust radius
        if isinstance(channel_set.montage, str):
            if channel_set.montage == '10-05':
                M = 345
            elif channel_set.montage == '10-10':
                M = 71
            elif channel_set.montage == '10-20':
                M = 21
        elif isinstance(channel_set.montage, dict) or channel_set.montage\
                is None:
            M = channel_set.n_cha
        percentage = len(channel_set.channels) * (0.25 / (M - 2)) + \
                     0.25 * ((M - 4) / (M - 2))
        min_dist = min_dist * percentage

    else:
        min_dist = chcontour_radius

    # Plot a contour around electrodes
    if plot_contour_ch:
        for ch_idx in range(len(channel_set.channels)):
            axes.add_patch(plt.Circle(
                (ch_x[ch_idx], ch_y[ch_idx]), radius=min_dist,
                facecolor='#ffffff', edgecolor=None, alpha=0.4, zorder=10))

    # Plotting the electrodes
    if plot_channels:
        axes.scatter(ch_x, ch_y, linewidth*3.5, facecolors='w', edgecolors='k',
                     zorder=10)

    if plot_clabels:
        for t in range(len(channel_set.channels)):
            axes.text(ch_x[t] + 0.01, ch_y[t] - 0.85 * min_dist,
                      channel_set.channels[t]['label'], fontsize=linewidth*2,
                      color='w',
                      zorder=11)

    # Last considerations
    plot_lim = max(head_radius + 0.2, np.max(np.sqrt(ch_x**2 + ch_y**2)) + 0.2)
    axes.set_xlim([-plot_lim, plot_lim])
    axes.set_ylim([-plot_lim, plot_lim])
    axes.set_aspect('equal', 'box')
    axes.axis('off')
    if fig is not None:
        # fig = plt.gcf()
        fig.patch.set_alpha(background)  # Set transparent background
        fig.tight_layout()
    if show is True:
        plt.show(dpi=400)

    return fig, axes


def ear_rho(ear_theta, ellipse_a, ellipse_b):
    """  This function computes the ear coordinates according to an ellipse.
    """
    d1 = np.power(np.cos(ear_theta), 2) / np.power(ellipse_a, 2)
    d2 = np.power(np.sin(ear_theta), 2) / np.power(ellipse_b, 2)
    return 1 / np.sqrt(d1 + d2)


def pol2cart(rho, phi):
    """This function converts polar coordinates to cartesian coordinates.

    Parameters
    ----------
    rho:     Array of radii
    phi:     Array of angles
    """
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return x, y


def compute_nearest_values(coor_add, coor_neigh, val_neigh, k):
    """ This function computes the mean values of the k-nearest neighbors.

    Parameters
    ----------
    coor_add:    XY coordinates of the virtual electrodes.
    coor_neigh:  XY coordinates of the real electrodes.
    val_neigh:   Values of the real electrodes.
    k:           Number of neighbors to consider.
    """
    add_val = np.empty((len(coor_add), 1))
    L = len(coor_add)

    for i in range(L):
        # Distances between the added electrode and the original ones
        target = coor_add[i, :] * np.ones((len(coor_neigh), 2))
        d = np.sqrt(np.sum(np.power(target - coor_neigh, 2), axis=1))

        # K-nearest neighbors
        idx = np.argsort(d)
        sel_idx = idx[1:1 + k]

        # Final value as the mean value of the k-nearest neighbors
        add_val[i] = np.mean(val_neigh[0, sel_idx])
    return add_val

def get_cartesian_coordinates(channel_set):
    # Restructure the channels list to treat it more easily
    if channel_set.coord_system == 'spherical':
        radius = list()
        theta = list()
        for c in channel_set.channels:
            try:
                radius.append(c['r'])
                theta.append(c['theta'])
            except KeyError as e:
                raise UnlocatedChannel(c)
        radius, theta = np.array(radius),np.array(theta)
        ch_x, ch_y = pol2cart(radius, theta)
    else:
        ch_x, ch_y = list(),list()
        for c in channel_set.channels:
            try:
                ch_x.append(c['x'])
                ch_y.append(c['y'])
            except KeyError as e:
                raise UnlocatedChannel(c)
        ch_x,ch_y = np.array(ch_x),np.array(ch_y)
    return ch_x, ch_y

if __name__ == "__main__":
    """ Example of use: """
    from medusa.meeg.meeg import EEGChannelSet

    # Set channel set
    channel_set = EEGChannelSet()
    channel_set.set_standard_montage(
        l_cha=['F3','F7','FZ', 'F4','F8', 'FCZ','C3', 'CZ', 'C4','CPZ', 'P3',
               'PZ', 'P4','PO7','POZ','PO8'],
        standard='10-10')

    # Plot topographic plot
    plt.figure()
    dummy_values_topo = np.arange(len(channel_set.channels))
    plot_topography(channel_set, dummy_values_topo, plot_clabels=True,
                    plot_contour_ch=True, plot_extra=0.1,
                    plot_skin_in_color=True,cmap='plasma')

    # Plot connectivity plot
    plt.figure()
    dummy_values_conn = np.random.randn(16,16)
    plot_connectivity(channel_set,dummy_values_conn)



import tensorflow as tf
import numpy as np

def __clustering_gpu(W):
    """
    Calculates the ClC of echa node of the graph contained in the matrix W using GPU

    It is calculated as follows:
              ___   ___  
              \     \    Wik * Wli * Wkl
              /__   /__  
              k~=i (l~=k & l~=i)
    ClC = -----------------------------
              ___   ___  
              \     \    Wik * Wil 
              /__   /__  
              k~=i (l~=k & l~=i)

    C.J. Stam, et ál. (2009). "Graph theoretical analysis of 
    magnetoencephalographic functional connectivity in Alzheimer’s 
    disease". Brain, 132:213–224.

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    nodal_clc : numpy array
        Nodal clustering coefficient
        
    """

    A = tf.where(W>0,1,0)
    S = tf.math.add(tf.math.pow(W, 1/3), tf.math.pow(tf.transpose(W), 1/3))
    K = tf.math.reduce_sum(tf.math.add(A,tf.transpose(A)),axis=1)
    cyc_3 = tf.math.divide(tf.linalg.tensor_diag_part(tf.linalg.matmul(tf.linalg.matmul(S,S),S)),2)
    K = tf.where(cyc_3==0,np.inf,tf.cast(K,tf.float32))
    f1 = tf.math.subtract(K,1)
    f2 = tf.math.multiply(2,tf.linalg.tensor_diag_part(tf.linalg.matmul(A,A)))
    f3 = tf.math.multiply(K,f1)
    CYC_3 = tf.math.subtract(f3,tf.cast(f2,tf.float32))
    nodal_clc = tf.math.divide(tf.cast(cyc_3, tf.float32),CYC_3)
        
    return nodal_clc


def __clustering_cpu(W):
    """
    Calculates the ClC of echa node of the graph contained in the matrix W using CPU

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    nodal_clc : numpy array
        Nodal clustering coefficient
        
    """
    
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')     

    A = np.ones(W.shape)
    A[W<=0] = 0
    S = np.power(W, 1/3) + np.power(np.transpose(W), 1/3)
    K = np.sum(A + np.transpose(A),axis=1)
    cyc_3 = np.diag(np.linalg.matrix_power(S,3)) / 2
    K[cyc_3 == 0] = np.inf
    f1 = np.multiply(K,(K - 1))
    f2 = 2 * np.diag(np.linalg.matrix_power(A,2))
    CYC_3 = f1 - f2
    nodal_clc = cyc_3 / CYC_3
        
    return nodal_clc


def clustering_coefficient(W,mode):
    """
    Calculates the clustering coefficient.

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
    mode : string
        GPU or CPU
        
    Returns
    -------
    nodal_clc : numpy array
        Nodal clustering coefficient

    """
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
      
    if mode == 'CPU':
        nodal_clc = __clustering_cpu(W)
    elif mode == 'GPU':
        nodal_clc = __clustering_gpu(W)
    else:
        raise ValueError('Unknown mode')
        
    return nodal_clc



"""
In this module you will find useful functions to apply optimized amplitude-based
connectivity metrics. Enjoy!

@authors: Víctor Rodríguez-González and Diego Marcos-Martínez
"""

# Built-in imports
import warnings

# External imports
import numpy as np
from scipy import stats as sp_stats
import tensorflow as tf
from tensorflow_probability import stats as tfp_stats

# Medusa imports
import medusa.components
from medusa import signal_orthogonalization as orthogonalizate
from medusa.transforms import hilbert
from medusa import pearson_corr_matrix as corr
from medusa.utils import check_dimensions


def __aec_gpu(data):
    """ This method implements the amplitude envelope correlation using GPU.

    NOTE: See the orthogonalized version.

    Parameters
    ----------
    data : numpy.ndarray
        MEEG Signal. [n_epochs, n_samples, n_channels].

    Returns
    -------
    aec : numpy.ndarray
        aec-based connectivity matrix. [n_epochs, n_channels, n_channels].
    """
    # Error check
    if type(data) != np.ndarray:
        raise ValueError("Parameter data must be of type numpy.ndarray")
    # Set to correct dimensions
    data = check_dimensions(data)

    # AEC computation
    hilb = hilbert(data)
    envelope = tf.math.abs(hilb) 
    env = tf.math.log(tf.math.square(envelope))

    aec = tfp_stats.correlation(env,sample_axis=1)

    return aec.numpy()


def __aec_cpu(data):
    """ This method implements the amplitude envelope correlation using CPU.

    NOTE: See the orthogonalized version.

    Parameters
    ----------
    data : numpy.ndarray
        MEEG Signal. [n_epochs, n_samples, n_channels].

    Returns
    -------
    aec : numpy.ndarray
        aec-based connectivity matrix. [n_epochs, n_channels, n_channels].

    """
    # Error check
    if type(data) != np.ndarray:
        raise ValueError("Parameter data must be of type numpy.ndarray")
    # Set to correct dimensions
    data = check_dimensions(data)

    #  Variable initialization
    n_epo = data.shape[0]
    n_cha = data.shape[2]
    aec = np.empty((n_epo, n_cha, n_cha))
    aec[:] = np.nan

    # AEC computation
    hilb = hilbert(data)
    envelope = abs(hilb)
    env = np.log(envelope ** 2)

    # Concurrent calculation for more than one epoch
    w_threads = []
    for epoch in env:
        t = medusa.components.ThreadWithReturnValue(target= np.corrcoef,
                                                    args=(epoch,None,False,))
        w_threads.append(t)
        t.start()

    for epoch_idx, thread in enumerate(w_threads):
        aec[epoch_idx, :, :] = thread.join()

    return aec

def __aec_ort_gpu(data):
    """ This method implements the orthogonalized version of the amplitude
    envelope correlation using GPU. This orthogonalized version minimizes the
    spurious connectivity caused by common sources (zero-lag correlations).

    Parameters
    ----------
    data : numpy.ndarray
        MEEG Signal. [n_epochs, n_samples, n_channels].

    Returns
    -------
    aec : numpy.ndarray
        aec-based connectivity matrix. [n_epochs, n_channels, n_channels].

    """
    # Error check
    if type(data) != np.ndarray:
        raise ValueError("Parameter data must be of type numpy.ndarray")
    # Set to correct dimensions
    data = check_dimensions(data)

    # Variable initialization
    n_epo = data.shape[0]
    n_samp = data.shape[1]
    n_cha = data.shape[2]
    aec_ort = np.empty((n_epo,n_cha, n_cha))
    aec_ort[:] = np.nan
    
    # AEC Ort Calculation (CPU orthogonalization is much faster than GPU one)
    signal_ort = orthogonalizate.signal_orthogonalization_cpu(data, data)
    signal_ort_2 = tf.transpose(tf.reshape(tf.transpose(signal_ort,perm=[0,3,2,1]),
                                           (n_epo,n_cha*n_cha,
                                            n_samp)),perm=[0,2,1])

    hilb_1 = hilbert(signal_ort_2)
    envelope_1 = tf.math.abs(hilb_1) 
    env = tf.math.log(tf.math.square(envelope_1))

    # Concurrent calculation for more than one epoch
    w_threads = []
    for epoch in env:
        t = medusa.components.ThreadWithReturnValue(target=__aec_ort_comp_aux,
                                                    args=(epoch, n_cha, 'gpu',))
        w_threads.append(t)
        t.start()

    for epoch_idx, thread in enumerate(w_threads):
        aec_ort[epoch_idx, :, :] = thread.join()
        
    return aec_ort


def __aec_ort_cpu(data):
    """ This method implements the orthogonalized version of the amplitude
    envelope correlation using CPU. This orthogonalized version minimizes the
    spurious connectivity caused by common sources (zero-lag correlations).

    Parameters
    ----------
    data : numpy.ndarray
        MEEG Signal. [n_epochs, n_samples, n_channels].

    Returns
    -------
    aec : numpy.ndarray
        aec-based connectivity matrix. [n_epochs, n_channels, n_channels].

    """
    # Error check
    if type(data) != np.ndarray:
        raise ValueError("Parameter data must be of type numpy.ndarray")

    # Set to correct dimensions
    data = check_dimensions(data)

    # Variable initialization
    n_epo = data.shape[0]
    n_samp = data.shape[1]
    n_cha = data.shape[2]
    aec_ort = np.empty((n_epo, n_cha, n_cha))
    aec_ort[:] = np.nan

    # AEC Ort Calculation
    signal_ort = orthogonalizate.signal_orthogonalization_cpu(data, data)
    signal_ort_2 = np.transpose(np.reshape(np.transpose(signal_ort, (0, 3, 2, 1)),
                            (n_epo, n_cha * n_cha, n_samp)),(0,2,1))

    hilb_1 = hilbert(signal_ort_2)
    envelope_1 = np.abs(hilb_1)
    env = np.log(np.square(envelope_1))

    # Concurrent calculation for more than one epoch
    w_threads = []
    for epoch in env:
        t = medusa.components.ThreadWithReturnValue(target=__aec_ort_comp_aux,
                                                    args=(epoch,n_cha,'cpu',))
        w_threads.append(t)
        t.start()

    for epoch_idx, thread in enumerate(w_threads):
        aec_ort[epoch_idx,:,:] = thread.join()

    return aec_ort

def __aec_ort_comp_aux(env, n_cha, ctype='cpu'):
    """
    Auxiliary method that implements a function to compute the orthogonalized AEC.
    Parameters
    ----------
    env: numpy.ndarray
        Array with signal envelope. [n_epochs, n_samples, n_channels x n_channels].
    type: str
        Calculation type: 'cpu' or 'gpu'.
    Returns
    -------
    aec_ort: numpy.ndarray
        AEC orthogonalized connectivity matrix. [n_channels, n_channels].
    """
    # Note: Orthogonalize A regarding B is not the same as orthogonalize B regarding
    # A, so we average lower and upper triangular matrices to construct the
    # symmetric matrix required for Orthogonalized AEC

    if ctype == 'cpu':
        aec_tmp = np.corrcoef(env, rowvar=False)
        aec_tmp2 = np.transpose(
            np.reshape(
                np.transpose(aec_tmp),
                (int(aec_tmp.shape[0] * aec_tmp.shape[0] / n_cha), -1)
            )
        )
        idx = np.linspace(0, aec_tmp2.shape[1] - 1, n_cha).astype(np.int32)
        aec = aec_tmp2[:, idx]
        aec_upper = np.triu(np.squeeze(aec))
        aec_lower = np.transpose(np.tril(np.squeeze(aec)))
        aec_ort = (aec_upper + aec_lower) / 2
        aec_ort = abs(np.triu(aec_ort, 1) + np.transpose(aec_ort))
        return aec_ort

    elif ctype == 'gpu':
        aec_tmp = tfp_stats.correlation(env)
        aec_tmp2 = tf.transpose(
            tf.reshape(
                tf.transpose(aec_tmp),
                (tf.cast(aec_tmp.shape[0] * aec_tmp.shape[0] / n_cha,
                         tf.int32), -1)
            )
        )
        idx = tf.cast(tf.linspace(0, len(aec_tmp2[0]) - 1, n_cha), tf.int32)
        aec = tf.gather(aec_tmp2, idx, axis=1).numpy()
        aec_upper = tf.linalg.band_part(aec, 0, -1)
        aec_lower = tf.transpose(tf.linalg.band_part(aec, -1, 0))
        aec_ort = tf.math.divide(tf.math.add(aec_upper, aec_lower), 2)
        aux = tf.linalg.band_part(aec_ort, 0, -1) - tf.linalg.band_part(
            aec_ort, 0, 0)
        aec_ort = tf.math.abs(tf.math.add(aux, tf.transpose(aec_ort)))
        return aec_ort

def aec(data, ort=True):
    """ This method implements the amplitude envelope correlation (using GPU if
    available). Based on the "ort" param, the signals could be orthogonalized
    before the computation of the amplitude envelope correlation.

    REFERENCES:
    Liu, Z., Fukunaga, M., de Zwart, J. A., & Duyn, J. H. (2010).
    Large-scale spontaneous fluctuations and correlations in brain electrical
    activity observed with magnetoencephalography. Neuroimage, 51(1), 102-111.

    Hipp, J. F., Hawellek, D. J., Corbetta, M., Siegel, M., & Engel,
    A. K. (2012). Large-scale cortical correlation structure of spontaneous
    oscillatory activity. Nature neuroscience, 15(6), 884-890.

    O’Neill, G. C., Barratt, E. L., Hunt, B. A., Tewarie, P. K., & Brookes, M.
    J. (2015). Measuring electrophysiological connectivity by power envelope
    correlation: a technical review on MEG methods. Physics in Medicine &
    Biology, 60(21), R271.

    Parameters
    ----------
    data : numpy.ndarray
        MEEG Signal. Allowed dimensions: [n_epochs, n_samples, n_channels] and
        [n_samples, n_channels].
    ort : bool
        If True, the signals on "data" will be orthogonalized before the
        computation of the amplitude envelope correlation.

    Returns
    -------
    aec : numpy.ndarray
        aec-based connectivity matrix. [n_epochs, n_channels, n_channels].

    """
    from medusa import tensorflow_integration
    #  Error check
    if not np.issubdtype(data.dtype, np.number):
        raise ValueError('data matrix contains non-numeric values') 

    if not ort:
        if tensorflow_integration.check_tf_config(autoconfig=True):
            aec = __aec_gpu(data)
        else:
            aec = __aec_cpu(data)

    else:
        if tensorflow_integration.check_tf_config(autoconfig=True):
            aec = __aec_ort_gpu(data)
        else:
            aec = __aec_ort_cpu(data)

    return aec


def __iac_gpu(data):
    """ This method implements the instantaneous amplitude correlation using
    GPU.

    NOTE: See the orthogonalized version. In the original paper, the
    orthogonalized version was used

    Parameters
    ----------
    data : numpy.ndarray
        MEEG Signal. [n_epochs, n_samples, n_channels].

    Returns
    -------
    iac : numpy.ndarray
        iac-based connectivity matrix.
        [n_epochs, n_channels, n_channels, n_samples].


    """
    # Error check
    if type(data) != np.ndarray:
        raise ValueError("Parameter data must be of type numpy.ndarray")

    # Set to correct dimensions
    data = check_dimensions(data)

    #  Variable initialization
    n_epo = data.shape[0]
    n_samp = data.shape[1]
    n_cha = data.shape[2]

    # Z-score
    data = tf.math.divide(
        tf.math.subtract(data,
                         tf.math.reduce_mean(data, axis=1, keepdims=True)),
        tf.math.reduce_std(data, axis=1, keepdims=True))

    #  IAC computation
    hilb = hilbert(data)
    envelope = tf.math.abs(hilb)

    iac = tf.multiply(tf.tile(envelope, (1, 1, n_cha)),
                      tf.transpose(tf.reshape(
                          tf.transpose(tf.tile(envelope, (1, n_cha,1)),perm=[0,2,1]),
                          (n_epo, n_cha*n_cha, n_samp)), perm=[0,2,1])).numpy()
    iac = tf.reshape(tf.transpose(iac,perm=[0,2,1]),
                     (n_epo, n_cha, n_cha, n_samp)).numpy()

    # Set diagonal to 0
    diag_mask = tf.ones((n_cha, n_cha))
    diag_mask = tf.linalg.set_diag(diag_mask, tf.zeros(diag_mask.shape[0:-1]),
                                   name=None)
    iac = tf.multiply(iac, tf.repeat(tf.repeat(diag_mask[None,:, :, None], n_samp,
                                     axis=-1), n_epo,axis=0))

    return iac.numpy()

def __iac_cpu(data):
    """ This method implements the instantaneous amplitude correlation using
    CPU.

    NOTE: See the orthogonalized version. In the original paper, the
    orthogonalized version was used

    Parameters
    ----------
    data : numpy.ndarray
        MEEG Signal. [n_epochs, n_samples, n_channels].

    Returns
    -------
    iac : numpy.ndarray
        iac-based connectivity matrix.
        [n_epochs, n_channels, n_channels, n_samples].

    """
    # Error check
    if type(data) != np.ndarray:
        raise ValueError("Parameter data must be of type numpy.ndarray")

    # Set to correct dimensions
    data = check_dimensions(data)

    #  Variable initialization
    n_epo = data.shape[0]
    n_samp = data.shape[1]
    n_cha = data.shape[2]

    # IAC computation
    data = sp_stats.zscore(data, axis=1)

    hilb = hilbert(data)
    envelope = abs(hilb)
    iac = np.multiply(np.reshape(
        np.tile(envelope, (1, n_cha, 1)), (n_epo, n_samp, n_cha*n_cha), order='F'),
        np.tile(envelope, (1, 1, n_cha)))
    iac = np.reshape(np.transpose(iac,(0,2,1)), (n_epo, n_cha, n_cha, n_samp))

    # Set diagonal to 0
    diag_mask = np.ones((n_cha, n_cha))
    np.fill_diagonal(diag_mask, 0)
    iac = iac * np.repeat(np.repeat(diag_mask[None,:, :, None], n_samp, axis=-1),
                          n_epo,axis=0)

    return iac


def __iac_ort_gpu(data):
    """ This method implements the orthogonalized version of the instantaneous
    amplitude correlation using GPU. This orthogonalized version minimizes the
    spurious connectivity caused by common sources (zero-lag correlations).

    data : numpy.ndarray
        MEEG Signal. [n_epochs, n_samples, n_channels].

    Returns
    -------
    iac_ort : numpy.ndarray
        iac-based connectivity matrix.
        [n_epochs, n_channels, n_channels, n_samples].
    """
    # Error check
    if type(data) != np.ndarray:
        raise ValueError("Parameter data must be of type numpy.ndarray")

    # Set to correct dimensions
    data = check_dimensions(data)

    #  Variable initialization
    n_epo = data.shape[0]
    n_samp = data.shape[1]
    n_cha = data.shape[2]

    # IAC Ort Calculation (CPU orthogonalization is much faster than GPU one)
    # Z-score
    data = tf.math.divide(
        tf.math.subtract(data,
                         tf.math.reduce_mean(data, axis=1, keepdims=True)),
        tf.math.reduce_std(data, axis=1, keepdims=True))

    signal_ort = orthogonalizate.signal_orthogonalization_cpu(data.numpy(),
                                                              data.numpy())

    signal_ort_2 = tf.transpose(tf.reshape(tf.transpose(signal_ort, perm=
                                                        [0,3,2,1]),
                                           (n_epo, n_cha * n_cha,n_samp)),
                                perm = [0,2,1])

    hilb_1 = hilbert(signal_ort_2)
    envelope = tf.math.abs(hilb_1)

    iac = tf.multiply(tf.tile(envelope, (1, 1, n_cha**2)),
                      tf.transpose(tf.reshape(
                          tf.transpose(tf.tile(envelope, (1, n_cha**2, 1)), perm=
                                       [0,2,1]),
                          (n_epo,(n_cha*n_cha)**2, n_samp)),perm=[0,2,1]))
    iac = tf.reshape(tf.transpose(iac,perm=[0,2,1]), (n_epo, n_cha**2, n_cha**2,
                                                      n_samp))

    iac_tmp2 = tf.transpose(
        tf.reshape(
            tf.transpose(iac, (0,2, 1, 3)),
            (n_epo,int(iac.shape[1] * iac.shape[1] / n_cha), -1, iac.shape[3])
        ), (0,2, 1, 3)
    )
    idx = tf.cast(tf.linspace(0, iac_tmp2.shape[2]-1, n_cha), dtype=tf.int32)
    iac = tf.gather(iac_tmp2, idx, axis=2)

    # Orthogonalize A regarding B is not the same as orthogonalize B regarding
    # A, so we average lower and upper triangular matrices to construct the
    # symmetric matrix required for Orthogonalized iac

    iac_upper = tf.linalg.band_part(tf.transpose(iac, (0,3, 1, 2)), 0, -1)
    iac_lower = tf.transpose(tf.linalg.band_part(tf.transpose(iac, (0,3, 1, 2)),
                                                 -1, 0), (0,1, 3, 2))
    iac_ort = tf.math.divide(tf.math.add(iac_upper, iac_lower), 2)
    aux = tf.linalg.band_part(iac_ort, 0, -1) - tf.linalg.band_part(iac_ort, 0,
                                                                    0)
    iac_ort = tf.math.abs(tf.math.add(aux, tf.transpose(aux, (0,1, 3, 2))))

    return tf.transpose(iac_ort, (0, 2, 3, 1)).numpy()

def __iac_ort_cpu(data):
    """ This method implements the orthogonalized version of the instantaneous
    amplitude correlation using CPU. This orthogonalized version minimizes the
    spurious connectivity caused by common sources (zero-lag correlations).

    Parameters
    ----------
    data : numpy.ndarray
        MEEG Signal. [n_epochs, n_samples, n_channels].

    Returns
    -------
    iac_ort : numpy.ndarray
        iac-based connectivity matrix.
        [n_epochs, n_channels, n_channels, n_samples].
    """
    # Error check
    if type(data) != np.ndarray:
        raise ValueError("Parameter data must be of type numpy.ndarray")

    # Set to correct dimensions
    data = check_dimensions(data)

    #  Variable initialization
    n_epo = data.shape[0]
    n_samp = data.shape[1]
    n_cha = data.shape[2]

    # AEC Ort Calculation
    data = sp_stats.zscore(data, axis=1)

    signal_ort = orthogonalizate.signal_orthogonalization_cpu(data, data)
    signal_ort_2 = np.transpose(
        np.reshape(np.transpose(signal_ort, (0, 3, 2, 1)),
                   (n_epo, n_cha * n_cha, n_samp)), (0, 2, 1))

    hilb_1 = hilbert(signal_ort_2)
    envelope_1 = np.abs(hilb_1)

    iac = np.multiply(np.reshape(np.tile(
        envelope_1, (1, n_cha**2, 1)), (n_epo, n_samp, n_cha**2*n_cha**2),
        order='F'), np.tile(envelope_1, (1, 1, n_cha**2)))
    iac = np.reshape(np.transpose(iac,[0,2,1]), (n_epo,n_cha**2, n_cha**2, n_samp))
    iac_tmp2 = np.transpose(
        np.reshape(
            np.transpose(iac, (0,2, 1, 3)),
            (n_epo,int(iac.shape[1] * iac.shape[1] / n_cha), -1, n_samp)
        ), (0,2, 1, 3)
    )
    idx = np.linspace(0, iac_tmp2.shape[2]-1, n_cha).astype(np.int32)
    iac = iac_tmp2[:,:, idx, :]

    # Orthogonalize A regarding B is not the same as orthogonalize B regarding
    # A, so we average lower and upper triangular matrices to construct the
    # symmetric matrix required for Orthogonalized AEC

    iac_upper = np.triu(np.transpose(iac, (0,3, 1, 2)), k=1)
    iac_lower = np.transpose(np.tril(np.transpose(iac, (0,3, 1, 2)), k=-1), (0,1, 3,
                                                                           2))
    iac_ort = (iac_upper + iac_lower) / 2
    iac_ort = abs(np.triu(iac_ort, k=1) + np.transpose(iac_ort, (0,1, 3, 2)))

    return np.transpose(iac_ort, (0,2, 3, 1))


def iac(data, ort=True):
    """ This method implements the instantaneous amplitude correlation (using
    GPU if available). Based on the "ort" param, the signals could be
    orthogonalized before the computation of the amplitude envelope correlation.

    REFERENCES:
    Tewarie, P., Liuzzi, L., O'Neill, G. C., Quinn, A. J., Griffa,
    A., Woolrich, M. W., ... & Brookes, M. J. (2019). Tracking dynamic brain
    networks using high temporal resolution MEG measures of functional
    connectivity. Neuroimage, 200, 38-50.

    O’Neill, G. C., Barratt, E. L., Hunt, B. A., Tewarie, P. K., & Brookes, M.
    J. (2015). Measuring electrophysiological connectivity by power envelope
    correlation: a technical review on MEG methods. Physics in Medicine &
    Biology, 60(21), R271.

    Parameters
    ----------
    data : numpy.ndarray
        MEEG Signal. Allowed dimensions: [n_epochs, n_samples, n_channels] and
        [n_samples, n_channels].
    ort : bool
        If True, the signals on "data" will be orthogonalized before the
        computation of the instantaneous amplitude correlation.

    Returns
    -------
    iac : numpy 2D square matrix
        iac-based connectivity matrix.
        [n_epochs, n_channels, n_channels, n_samples].

    """
    from medusa import tensorflow_integration

    #  Error check
    if not np.issubdtype(data.dtype, np.number):
        raise ValueError('data matrix contains non-numeric values')

    if not ort:
        if tensorflow_integration.check_tf_config(autoconfig=True):
            iac = __iac_gpu(data)
        else:
            iac = __iac_cpu(data)

    else:
        if tensorflow_integration.check_tf_config(autoconfig=True):
            iac = __iac_ort_gpu(data)
        else:
            iac = __iac_ort_cpu(data)

    return iac



import numpy as np


def eigen_centrality(W):
    """
    Calculates the eigenvector centrality, which is a centrality measure based
    on the adjacency matrix eigenvectors
    
    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
    mode : string
        GPU or CPU
        
    Returns
    -------
    nodal_eig : numpy array
        Nodal eigenvector centrality

    """
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
        
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')        
      
    N = W.shape[0] # Number of nodes
    
    D,V = np.linalg.eig(W)
    
    idx = np.where(D == max(D))[0]
    
    nodal_eig = abs(V[:,idx])
    
        
    return nodal_eig

# import scipy.io as rmat

# data = rmat.loadmat('D:/OneDrive - Universidad de Valladolid/Scripts/testPython/graphTest.mat')
# W = data['W']
# W = np.squeeze(W)
# aa = eigen_centrality(W)


import numpy as np
from medusa.graph_theory import degree,modularity

def participation_coefficient(W):
    """
    Calculates the participation coefficient, which is the diversity of the 
    between-module connections of each node

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    nodal_part : numpy array
        Nodal participation coefficient.

    """
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
        
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')

    N = W.shape[0] # Graph size       
        
    _,community = modularity.modularity(W, 1) # Community vector
    
    deg = degree.degree(W, 'CPU') # Degree of each node
    
    Gc = np.matmul(np.where(deg != 0, 1,0)[:,np.newaxis],community.T) # Multiplying each graph row by its community vector
    
    # Add iteratively the participation of each node in each group
    Kc = np.zeros((N,1))
    for ii in range(1,np.amax(community).astype(int)+1):
        Kc = Kc + (np.sum(W * np.where(Gc==ii,1,0),axis=1)**2)[:,np.newaxis]
        
    # Guimera and Amaral normalisation
    nodal_part = np.ones((N,1)) - Kc / (deg.T**2)[:,np.newaxis]
    nodal_part[deg == 0] = 0
    
    return nodal_part

# import scipy.io as rmat

# data = rmat.loadmat('D:/OneDrive - Universidad de Valladolid/Scripts/testPython/graphTest.mat')
# W = data['W']
# aa = participation_coefficient(W)



# from medusa.remote.server.configure import *
# configure(sql_host='localhost', sql_user='root', sql_passwd='', db_folder='E://BBDD/medusa-server-databases/')

import sys, os
from medusa.bci.remote.erp_spellers_routes import erp_spellers_routing

from flask import Flask

# Parameters
if len(sys.argv) == 1:
    port = 50000
else:
    port = sys.argv[1]

app = Flask("medusa-server")
app.root_path = os.path.dirname(os.path.abspath(__file__))
app.register_blueprint(erp_spellers_routing)
app.run(host="0.0.0.0", port=port)



import numpy as np

def surrogate_graph(W):
    """
    Calculates the graph degree.

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
        
    Returns
    -------
    surrog_W : numpy array
        Surrogated graph.      

    """
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
        
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')        
      
    surrog_matrix = np.zeros((W.shape[0],W.shape[1]))
    idx_up = np.argwhere(np.triu(W, k=1))
    val_up = W[idx_up[:,0],idx_up[:,1]]
    val_up_surrog = val_up[np.random.permutation(val_up.shape[0])]
    surrog_matrix[idx_up[:,0],idx_up[:,1]] = val_up_surrog
    surrog_matrix = surrog_matrix + np.transpose(surrog_matrix)

    return surrog_matrix


import numpy as np
from medusa import components


class EMG(components.BiosignalData):
    """Electromiography (EMG) biosignal data class.
    """

    def __init__(self, times, signal, fs, channel_set, location=None, **kwargs):
        """EMG constructor

        Parameters
        ----------
        times : list or numpy.ndarray
            1D numpy array [n_samples]. Timestamps of each sample. If they are
            not available, generate them
            artificially. Nevertheless, all signals and events must have the
            same temporal origin
        signal : list or numpy.ndarray
            2D numpy array [n_samples x n_channels]. EMG samples (the units
            should be defined using kwargs)
        fs : int or float
            Sample rate of the recording.
        channel_set : list or Object
            Channel information
        location : string
            Location of the recording (e.g., quadriceps)
        kwargs: kwargs
            Key-value arguments to be saved in the class. This general class
            does not check anything
        """

        # Standard attributes
        self.times = times
        self.signal = signal
        self.fs = fs
        self.channel_set = channel_set
        self.location = location

        # Set the specified arguments
        for key, value in kwargs.items():
            setattr(self, key, value)

    def to_serializable_obj(self):
        rec_dict = self.__dict__
        for key in rec_dict.keys():
            if type(rec_dict[key]) == np.ndarray:
                rec_dict[key] = rec_dict[key].tolist()
        return rec_dict

    @classmethod
    def from_serializable_obj(cls, dict_data):
        return cls(**dict_data)


import numpy as np
import tensorflow as tf
import scipy.special as sps
import tensorflow_probability as tfp

def __divergency_cpu(W):
    """
    Calculates the graph divergency. Its the entropic distance (as euclidean
    distance) between a uniform weight distribution (random graph) and the 
    network under study using CPU
    
    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.

    Returns
    -------
    D : numpy array
        Network divergency.
        
    """    
    N = W.shape[0]
    ind = np.argwhere(np.triu(np.ones((N,N)),k=1))
    W_wei = np.sum(W[ind[:,0],ind[:,1]]) # Tri Up indices
    
    if not W.all() == 0:
        p = W[ind[:,0],ind[:,1]] / W_wei # Weights normalising
    
        p_sort = np.sort(p,axis=0)[::-1] # Weight sorting (descending)
        
        p_equi_sort = 1 / len(p_sort) * np.ones((1,len(p_sort))) # Full network is the
        #equilibrium network (same weight for all the edges)
        
        # Normalised euclidean distance
        c = sps.comb(N,2) # Number of connections without autoloops
        D = np.sqrt(c / (c - 1)) * np.sqrt(np.nansum(
                np.power(np.squeeze(np.asarray(p_equi_sort))-np.squeeze(np.asarray(p_sort)),2)))
    else: # Empty graph
        D = 0 
        
    return D


def __graph_entropy_cpu(W):
    """
    Calculates the Shannon entropy of a weighted graph using CPU
    
    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.

    Returns
    -------
    H : numpy array
        Network entropy.
        
    """     
    N = W.shape[0]
    ind = np.argwhere(np.triu(np.ones((N,N)),k=1))
    W_wei = np.sum(W[ind[:,0],ind[:,1]]) # Tri Up indices

    if not W.all() == 0:
        p = W[ind[:,0],ind[:,1]] / W_wei # Weights normalising

        H = -1 * np.sum(p[p > 0] * np.log2(p[p>0])) # Shannon entropy calculation

        # Normalising H to range [0 1]
        K = np.log2(p.shape[0])
        H = H/K
    else: # Empty or full graph
        H = 1

    return H         
    

def __complexity_cpu(W):
    """
    Calculates the Shannon Graph Complexity of a graph that node i belong to 
    one of the network shortest paths using CPU
    
    Note: W must be converted to a connection-length matrix. It is common to 
    obtain it via mapping from weight to length. BC is normalised to the 
    range [0 1] as BC/[(N-1)(N-2)]
        
    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.

    Returns
    -------
    C : numpy array
        Network complexity.
        
    """
    tmp_1 = __divergency_cpu(W)
    tmp_2 = __graph_entropy_cpu(W)
    C = tmp_1 * tmp_2
                
    return C

def __aux_divergency(W,N,ind,W_wei):
    p = tf.math.divide(tf.gather_nd(W,tf.cast(ind,tf.int32)),W_wei) # Weights normalising

    p_sort = tf.sort(p,axis=0,direction='DESCENDING') # Weight sorting (descending)
    
    p_equi_sort = tf.math.multiply(tf.math.divide(1,p_sort.shape[0]),tf.ones((1,p_sort.shape[0]),dtype=tf.float64)) # Full network is the
    #equilibrium network (same weight for all the edges)
    
    # Normalised euclidean distance
    N = tf.cast(N,tf.float32)
    c1 = tfp.distributions.Binomial(total_count=N, probs=0.5) # Binomial distr.
    c2 = tf.math.multiply(tf.math.pow(0.5,2),tf.math.pow((1-0.5),tf.math.subtract(N,tf.constant(2,dtype=tf.float32))))
    c = tf.math.round(tf.math.divide(c1.prob(2),c2)) # N choose 2
    f1 = tf.math.sqrt(tf.math.divide(c,tf.math.subtract(c,1.0)))
    f2 = tf.math.pow(tf.math.subtract(tf.reshape(p_equi_sort,[-1]),tf.reshape(p_sort,[-1])),2)
    f2 = tf.math.sqrt(tf.experimental.numpy.nansum(f2,dtype=tf.float32))
    D = tf.math.multiply(f1,f2)
    
    return D
        
        
def __divergency_gpu(W):
    """
    Calculates the graph divergency. Its the entropic distance (as euclidean
    distance) between a uniform weight distribution (random graph) and the 
    network under study using GPU
    
    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.

    Returns
    -------
    D : numpy array
        Network divergency.
        
    """    
    W = tf.convert_to_tensor(W)
    N = W.shape[0]
    
    ind = tf.where(tf.linalg.band_part(tf.ones((N,N)), 0, -1) - tf.linalg.band_part(tf.ones((N,N)), 0, 0))
    W_wei = tf.reduce_sum(tf.gather_nd(W,ind)) # Tri Up indices
    
    D = tf.cond(tf.reduce_sum(W)!=0, lambda: __aux_divergency(W,N,ind,W_wei),lambda: tf.constant(0))
        
    return D


def __aux_entropy(W,ind,W_wei):
    p = tf.math.divide(tf.gather_nd(W,tf.cast(ind,tf.int32)),W_wei) # Weights normalising
    
    H1 = tf.constant(-1,dtype=tf.float64)
    H2 = tf.gather_nd(p,tf.where(p>0))
    H3 = tf.experimental.numpy.log2(tf.transpose(tf.gather_nd(p,tf.where(p>0))))
    H4 = tf.reduce_sum(tf.math.multiply(H2,H3))

    H = tf.math.multiply(H1,H4)

    K = tf.experimental.numpy.log2(p.shape[0])
    
    H = tf.math.divide(H,K)
    return H

def __graph_entropy_gpu(W):
    """
    Calculates the Shannon entropy of a weighted graph using GPU
    
    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.

    Returns
    -------
    H : numpy array
        Network entropy.
        
    """     
    W = tf.convert_to_tensor(W)
    N = W.shape[0]
    
    ind = tf.where(tf.linalg.band_part(tf.ones((N,N)), 0, -1) - tf.linalg.band_part(tf.ones((N,N)), 0, 0))
    W_wei = tf.reduce_sum(tf.gather_nd(W,ind)) # Tri Up indices
    
    H = tf.cond(tf.reduce_sum(W)!=0, lambda: __aux_entropy(W,ind,W_wei),lambda: tf.constant(1))

    return H         
    

def __complexity_gpu(W):
    """
    Calculates the Shannon Graph Complexity of a graph that node i belong to 
    one of the network shortest paths using GPU
    
    Note: W must be converted to a connection-length matrix. It is common to 
    obtain it via mapping from weight to length. BC is normalised to the 
    range [0 1] as BC/[(N-1)(N-2)]
        
    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.

    Returns
    -------
    C : numpy array
        Network complexity.
        
    """
    tmp_1 = __divergency_gpu(W)
    tmp_2 = __graph_entropy_gpu(W)
    C = tf.math.multiply(tmp_1,tf.cast(tmp_2,tf.float32))
                
    return C


def complexity(W,mode):
    """
    Calculates the graph complexity.

    Parameters
    ----------
    W : numpy 2D matrix
        Graph matrix. ChannelsXChannels.
    mode : string
        GPU or CPU
        
    Returns
    -------
    global_comp : numpy array
        Global complexity.

    """
    if W.shape[0] is not W.shape[1]:
        raise ValueError('W matrix must be square')
        
    if not np.issubdtype(W.dtype, np.number):
        raise ValueError('W matrix contains non-numeric values')        
      
    if mode == 'CPU':
        global_comp = __complexity_cpu(W)
    elif mode == 'GPU':
        global_comp = __complexity_gpu(W)
    else:
        raise ValueError('Unknown mode')
        
    return global_comp
 


import functools, time


def perf_analysis(func):
    """Returns the run time of the decorated function"""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        return value, {'run_time': run_time}
    return wrapper
